Attribute VB_Name = "modC3dReader"
'
' ****************************************************************
' ****************************************************************
' ****************************************************************
' This program is free software: you can redistribute it and/or
' modify it under the terms of the GNU General Public License
' version 3 as published my the Free Software Foundation:
' http://www.gnu.org/licenses/gpl.html
'
' This program is distributed in the hope that it will be useful,
' but WITHOUT ANY WARRANTY; without even the implied warranty of
' MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
' GNU General Public License for more details.
'
' This software was originally written by Elroy Sullivan, PhD in
' cooperation with Shriners Hospitals for Children, Houston.
' Dr. Sullivan can be reached at elroysullivan@gmail.com.
' ****************************************************************
' ****************************************************************
' ****************************************************************
'
'
' General notes on C3D files:
'
'   Video and analog data will always be in the same format (either integer or single).
'
'   The overall structure of a C3D file is:
'       A single 512 byte header section.
'       A "parameter" section with one or more 512 byte blocks.
'       A 3D point & analog data section of one or more 512 byte blocks.
'           The timeslices of 3D point data are often called frames (or 3D frames).
'           There may be multiple timeslices of analog data per timeslice of 3D data.
'               These may be called sub-frames or analog timeslices.
'               These analog sub-frames are all contained within the frames of the 3D data.
'               As per specifications, the analog framerate MUST be an exact multiple of the 3D data framerate.
'
'   These 512 byte sections are often called "blocks".
'
'   After all the header record and parameter blocks, the data records are stored sequentially by 3D data frames.
'       These frames are the timeslices of the 3D data.
'       All the 3D point data is in each of these frames.
'       Analog data is also stored in these frames (for all of the analog channels).
'           If analog data is captured at a faster rate than 3D data,
'           then there will be multiple timeslices of analog data in each frame.
'           For instance, analog data may be captured 50 times faster than 3D data,
'           Therefore, each C3D frame will contain one timeslice of 3D data and
'           50 timeslices of analog data.  The order through the analog channels (labels) is
'           repeated as many times as is needed to capture all the analog data.
'       The "FrameRecordSize" variable calculated in the procedures below explains much of this.
'       Essentially:
'           FrameRecordSize = NumberOf3dLabels * 4 * DataSize + AnalogChannelsCount * AnalogSamplesPerChannel * DataSize
'               DataSize = 2 for integers and 4 for singles.
'               NumberOf3dLabels is the number of 3D joints identified in the file.
'                   The * 4 is the three planes plus a pad (containing camera bit information).
'               AnalogChannelsCount is just the number of analog channels (labels) fed into the C3D file from analog equipment.
'               AnalogSamplesPerChannel is the number of analog timeslices per each 3D timeslice.
'
'
Option Explicit
'
Public Const GcdFrameCount = 51
'
Public Type GcdFrameType
    X As Single
    Y As Single
    z As Single
    Time As Single
End Type
'
Public Type GcdSingleValueMeasuresType
    SagittalDimension As Long   ' 1=X or 2=Y for determining the sagittal plane.
    Side As FootSideEnum
    StrideNumberForSide As Long
    StrideTime As Single
    Cadence As Single
    Speed As Single
    StepTimePct As Single
    OppositeFootOffPct As Single
    OppositeFootStrikePct As Single
    SameFootOffPct As Single
    SingleSupportPct As Single
    DoubleSupportPct As Single
    StrideLength As Single
    StepLength As Single
    VideoRate As Single
End Type
'
Public Type GcdFileSideType ' Actually, a separate variable is used for left and right sides.
    StrideTime As Single
    Cadence As Single
    StepTime As Single
    OppositeFootOff As Single
    OppositeFootContact As Single
    FootOff As Single
    SingleSupport As Single
    DoubleSupport As Single
    StrideLength As Single
    StepLength As Single
    Speed As Single
    VideoRate As Single
    Comments As String ' Not very useful, contains pt# and walks used (by VCM).
    AdditionalComments As String ' Previous comments used by this program.
    '
    WristFlexExt() As Single
    WristUlRadDeviation() As Single
    ElbowFlexExt() As Single
    ElbowProSup() As Single
    HeadFlexExt() As Single
    HeadLateralBend() As Single
    HeadRotation() As Single
    ShoulderFlexExt() As Single
    ShoulderAbAdduct() As Single
    ShoulderRotation() As Single
    TrunkFlexExt() As Single
    TrunkLateralBend() As Single
    TrunkRotation() As Single
    TrunkFlexExtAbs() As Single
    TrunkLateralBendAbs() As Single
    TrunkRotationAbs() As Single
    PelvicTiltRab() As Single
    PelvicObliquityRab() As Single
    PelvicRotationRab() As Single
    PelvicTiltBaker() As Single
    PelvicObliquityBaker() As Single
    PelvicRotationBaker() As Single
    AnkleFlexExtAbs() As Single
    '
    PelvicTilt() As Single
    PelvicObliquity() As Single
    PelvicRotation() As Single
    HipFlexExt() As Single
    HipAbAdduct() As Single
    HipRotation() As Single
    KneeFlexExt() As Single
    KneeValgVar() As Single
    KneeRotation() As Single
    DorsiPlanFlex() As Single
    FootRotation() As Single
    FootProgression() As Single
    '
    HipFlexExtMoment() As Single
    HipAbAdductMoment() As Single
    HipRotationMoment() As Single
    KneeFlexExtMoment() As Single
    KneeValgVarMoment() As Single
    KneeRotationMoment() As Single
    DorsiPlanFlexMoment() As Single
    FootAbAdductMoment() As Single
    FootRotationMoment() As Single
    '
    HipPower() As Single
    HipFlexExtPower() As Single
    HipAbAdductPower() As Single
    HipRotationPower() As Single
    KneePower() As Single
    KneeFlexExtPower() As Single
    KneeValgVarPower() As Single
    KneeRotationPower() As Single
    AnklePower() As Single
    DorsiPlanFlexPower() As Single
    AnkleAbAdductPower() As Single
    AnkleRotationPower() As Single
End Type
'
Private Type C3dStrikeOffEventFramesType
    ' This is used to save data on which frames have strike/off events.
    FrameNumber As Long
    StrideEventSide As FootSideEnum  ' 1 = Left, 2 = Right.
    StrideEventType As EventTypeEnum ' 1 = Foot Strike, 2 = Foot Off.
End Type
'
' These two are set up when the file is open, and are used to quickly identify valid strike/off events on point frame data.
Private StrikeOffEventFrames() As C3dStrikeOffEventFramesType
Private StrikeOffEventFramesCount As Long
'
' This one is just used to quickly blank out the strike/off events.
Private BlankStrikeOffEvents As StrikeOffEventsType
'
Public Enum EventTypeEnum
    FootStrike = 1
    FootOff = 2
    OtherEvent = 0
End Enum
#If False Then ' Intellisense fix.
    Public FootStrike
    Public FootOff
    Public OtherEvent
#End If
'
Public Enum FootSideEnum
    NoSide = 0
    LeftSide = 1
    RightSide = 2
End Enum
#If False Then ' Intellisense fix.
    Public NoSide
    Public LeftSide
    Public RightSide
#End If
'
Public Type StrikeOffStrideType
    ' This should only be used for VALID strike/off events.
    Side As FootSideEnum
    NumberForSide As Long
    '
    StartStrikeTime As Single
    OppositeFootOffTime As Single
    OppositeFootStrikeTime As Single
    SameFootOffTime As Single
    StopStrikeTime As Single
End Type
'
Public Type StrikeOffSingleEventType
    Side As FootSideEnum    ' 0 = no side, 1 = left, 2 = right.
    Type As EventTypeEnum   ' 1 = Strike, 2 = Off, 0 = All other (use label).
    Time As Single
End Type
'
Public Type StrikeOffEventsType
    TotalCount As Long                           ' This can also be used as a FOUND flag.
    '
    SidedStrikeOffCount As Long                  ' This is used EVEN IF EVENTS DON'T MAKE SENSE.
    SidedGeneralCount As Long                    ' This is used EVEN IF EVENTS DON'T MAKE SENSE.
    NonSidedCount As Long                        ' This is used EVEN IF EVENTS DON'T MAKE SENSE.
    '
    SidedStrikeOffEvents() As StrikeOffSingleEventType ' These are Foot-Strike or Foot-Off events with an identified side.
    SidedGeneralEvents() As StrikeOffSingleEventType   ' These have a side, but are not Foot-Strike or Foot-Off events.
    NonSidedEvents() As StrikeOffSingleEventType       ' These don't have a side.
    '
    ' The variables from here down are used to locate and normalize strides.
    ValidStrideCount As Long                     ' Use this as an "ANY FOUND" flag.
    ValidStrides() As StrikeOffStrideType              ' This is used to create the GCD files.
    ValidStrikeOffCount As Long
    ValidStrikeOffEvents() As StrikeOffSingleEventType ' This is used primarily to "TAG" the frames.  They have any initial and/or terminal foot-off events removed.
End Type
'
Public Const EmgLabelLength = 16 ' Labels are always 16 bytes long if written by NEXUS.
Public Const EmgDescLength = 50  ' Descriptions are always 50 bytes long if written by NEXUS.
Public Const EmgChannels = 16
'
' One caveat about these procedures is that a file MUST be opened to use them.
' The best approach for this is to REQUIRE that a program have an open C3D file before proceeding.
'
Private Type C3dHeaderType                           ' words bytes
    ParamBlockPointer As Byte                       '   1   1   Start block of Parameter records.
    FileID As Byte                                  '       2   Must be 80 decimal.
    PointLabelCount As Integer                      '   2   3   Number of different type of point data records.
    AnalogValuesPerFrame As Integer                 '   3   5   AnalogChannelsCount * AnalogSamplesPerChannel.  Total number of analog values for each point timeslice.
    FirstPointFrameNumber As Integer                '   4   7   ONE based.
    LastPointFrameNumber As Integer                 '   5   9
    MaxInterpolationGapAllowed As Integer           '   6  11
    ScaleFactor As Single                           '   7  13   If negative, data records are stored as SINGLE floating point.  Negative sign is removed after initial read.
    FirstDataBlockPointer As Integer                '   9  17   Start block of Point and Analog data records.  This should be treated as an UNSIGNED type byte integer.
    AnalogSamplesPerChannel As Integer              '  10  19   The number of analog timeslices for each point timeslice (per channel).
    PointFrameRate As Single                        '  11  21
    Reserved1 As String * 274                       '  13  25
    EventDataFlag As Integer                        ' 150 299   12345 decimal indicates presence.
    TimedEventsCount As Integer                     ' 151 301   Max of 18.
    Reserved2 As String * 2                         ' 152 303
    EventTimes(1 To 18) As Single                   ' 153 305
    EventDisplaySwitches(1 To 18) As Byte           ' 189 377   0=ON, 1=OFF.
    Reserved3 As String * 2                         ' 198 395
    EventLabels(1 To 18) As String * 4              ' 199 397
    Reserved4 As String * 44                        ' (235 to 256 words inclusive) or (469 to 512 bytes inclusive)
    '
    ' The following are not actually in the file.  They will be read, but will initially be garbage.
    ' If this is ever used to WRITE to a file, be sure these variables are taken out.
    AnalogChannelsCount As Single                   ' AnalogValuesPerFrame / AnalogSamplesPerChannel
    AnalogFrameRate As Single                       ' PointFrameRate * AnalogSamplesPerChannel
    SeekPtrForFirstC3dParam As Long
    SeekPtrForFirstFrameData As Long
    ProcessorType As Byte ' 84 = Intel; 85 = DEC; 86 = MIPS.
    DataStoredAsIntegerOrSingle As Byte ' 1 = Integer; 2 = Single.
    PointFrameCount As Long
    PrefixIsUsed As Boolean
    PrefixString As String
    ElapsedTimeOfWalk As Single
    WalkStartTime As Single
    WalkEndTime As Single
    TimeBetweenFrames As Single
End Type
'
Private Type C3dHeaderRwType
    ' This is exactly the same as the above except the calculated fields are removed.
    ' This one can be used for writing back to the file.
                                                    ' words bytes
    ParamBlockPointer As Byte                       '   1   1   Start block of Parameter records.
    FileID As Byte                                  '       2   Must be 80 decimal.
    PointLabelCount As Integer                      '   2   3   Number of different type of point data records.
    AnalogValuesPerFrame As Integer                 '   3   5   AnalogChannelsCount * AnalogSamplesPerChannels.
    FirstPointFrameNumber As Integer                '   4   7   ONE based.
    LastPointFrameNumber As Integer                 '   5   9
    MaxInterpolationGapAllowed As Integer           '   6  11
    ScaleFactor As Single                           '   7  13   If negative, data records are stored as SINGLE floating point.  Negative sign is removed after initial read.
    FirstDataBlockPointer As Integer                '   9  17   Start block of Point and Analog data records.  This should be treated as an UNSIGNED type byte integer.
    AnalogSamplesPerChannel As Integer              '  10  19
    PointFrameRate As Single                        '  11  21
    Reserved1 As String * 274                       '  13  25
    EventDataFlag As Integer                        ' 150 299   12345 decimal indicates presence.
    TimedEventsCount As Integer                     ' 151 301   Max of 18.
    Reserved2 As String * 2                         ' 152 303
    EventTimes(1 To 18) As Single                   ' 153 305
    EventDisplaySwitches(1 To 18) As Byte           ' 189 377   0=ON, 1=OFF.
    Reserved3 As String * 2                         ' 198 395
    EventLabels(1 To 18) As String * 4              ' 199 397
    Reserved4 As String * 44                        ' (235 to 256 words inclusive) or (469 to 512 bytes inclusive)
End Type
'
Public Type C3dParamHeaderType
    ParamBlockPointer As Byte ' Ignored in C3D files.
    RecordID As Byte ' Must be 80 decimal. (But ignored in C3D files.)
    ParamBlockCount As Byte ' Unsigned 0 to 255
    ProcessorType As Byte ' 84 = Intel; 85 = DEC; 86 = MIPS.
    ' Parameter data can start immediately following this header.
    ' But it must end on a 512 byte block boundary.
    '
    ' Do NOT add any variables NOT in the file to this structure.  Seek() is used after this structure is read.
End Type
'
Public Type C3dParamType
    ' This same structure is used for both groups and actual parameters.
    ' This structure is NOT used for direct file input because of the
    ' rather bazaar way in which these records are stored.  See the
    ' GetC3dParam function for further details.
    '
    ' These are used for both group and parameter records:
    NameLength As Byte              ' Length of Name string.
    ID As Byte                      ' Negative (>127) means it's a group label.  This can be used as a USED flag.
    Name As String                  ' This string MUST be set to the length of NameLength before this structure is used.
    OffsetForNext As Integer
    '
    ' These are used only for parameter type records:
    DataType As Byte                ' -1 (255) = char, 1 = byte, 2 = integer, 4 = single.
    OriginalDimensionCount As Byte  ' After the record is initially read, this isn't used.
    DimValue() As Byte              ' Not used if DimensionCount = 0 (but we actually set them to 1).
    TextLength As Byte              ' This is in the first DimValue() value when DataType = -1(255).
    Text() As String
    Bytes() As Byte
    Integers() As Integer
    Singles() As Single
    '
    ' These are used for both group and parameter records:
    DescriptionLength As Byte       ' Length of Description string.
    Description As String           ' This string MUST be set to the length of DescriptionLength before this variable is used for input.
    '
    ' Not in file structure at all:
    IsGroupOrParam As Byte          ' 1 = group, 2 = param.
    NextSeekPointer As Long         ' Applies to both group and param type records.  Set to ZERO when at the last record.
    Locked As Boolean               ' Similar to a read-only flag.
    ParamDataFound As Boolean       ' Used only on parameter type records.
End Type
'
' The following two are internal to this module for reading data records.
Private Type C3dPointFrameIntegerDataType
    ' This is used for files that store data as integers (DataStoredAsIntegerOrSingle = 1).
    X As Integer        ' Must be multiplied by scaling factor.
    Y As Integer        ' Must be multiplied by scaling factor.
    z As Integer        ' Must be multiplied by scaling factor.
    Reserved As Byte    ' Camera bits and residual are in here, but we won't need this.
End Type
Private Type C3dPointFrameSingleDataType
    ' This is used for files that store data as singles (DataStoredAsIntegerOrSingle = 2).
    X As Single
    Y As Single
    z As Single
    Reserved As Long    ' Camera bits and residual are in here, but we won't need this.
End Type
'
' This is the actual record that is returned when reading data.  It is not actually used for file I/O.
Public Type C3dPointFrameDataType
    X As Single
    Y As Single
    z As Single
    Time As Single
    SeekPointerToNextPointFrameData As Long ' This one does not go to zero when there are no more records.  The caller must know how many frames to get.
    SeekPointerToNextPointLabelData As Long ' This one does not go to zero when there are no more records.  The caller must know how many points to get.
    StrideEventType As Long                 ' 0 = no event, 1 = Foot Strike, 2 = Foot Off.  Used for strides that make sense from strike/off.
    StrideEventSide As Long                 ' 1 = Left, 2 = Right.  Only used if StrideEventType <> OtherEvent.
End Type
'
' This is the actual record that is returned when reading data.  It is not actually used for file I/O.
' It returns one frame of data for a specific analog channel (label).
Public Type C3dAnalogFrameDataType
    X() As Single                            ' The array accomodates the multiple analog timeslices per 3D time slice.
    Time As Single
    SeekPointerToNextFrameData As Long ' This one does not go to zero when there are no more records.  The caller must know how many frames to get.
    StrideEventType As Long                  ' 0 = no event, 1 = Foot Strike, 2 = Foot Off.  Used for strides that make sense from strike/off.
    StrideEventSide As Long                  ' 1 = Left, 2 = Right.  Only used if StrideEventType <> OtherEvent.
End Type
'
' The following structures are not used for getting records.
' They are used to organize data.
'
Public Type C3dGroupAndParamType
    C3dGroup As C3dParamType
    C3dParam As C3dParamType
End Type
'
Private Const C3dBlockSize As Long = 512&
Private Const C3dMaxDimensions As Byte = 7
'
' Public variables for use anywhere (THE C3dPointLabels is used in the C3dReader).
Public C3dPointLabels() As String ' I would put this in the header but it's a dynamic array.
Private C3DAnalogLabels() As String
Private C3DAnalogDescriptions() As String
Private C3dHeaderRw As C3dHeaderRwType
'
Public C3dFileSpec As String
Public EvtFileSpec As String
Public DtaFileSpec As String
Public C3dHeader As C3dHeaderType
'
' This is used all over the place, so let's leave it in memory.
Public StrikeOffEvents As StrikeOffEventsType
'
' Private variables for use only in this module.
Public C3dFileNumber As Long
Public DtaFileNumber As Long
' The EVT file is only temporarily opened so a handle is not needed.
'
Private Type DtaGraphTabType
    TheLabel As String
    TheColumn As Long ' ONE based.
End Type
Private DtaFileLine As String
Private DtaFrameNumber As Long
Private DtaGraphTabName() As DtaGraphTabType
Private DtaGraphPlotName() As String
Private DtaFirstFrameLine As Long ' The DATA file line on which the first frame is on.
Private DtaFileLinePtr As Long ' This is the line number that the file is on.
'
Private Type DtaPlotNamesType
    X As String ' Typically sagittal.
    Y As String ' Typically coronal.
    z As String ' Typically transverse.
End Type
'

Private Function DtaGetGraphPlotNames(sSideLtr As String, sGraphTab As String, iMoreSpecific As Long) As DtaPlotNamesType
    ' This is case sensitive.
    ' It returns empty strings if not found.
    ' See the code below for meanings of iMoreSpecific.
    Dim PlotNames As DtaPlotNamesType
    '
    Select Case sGraphTab
    Case "Trunk/Pelvis"
        Select Case iMoreSpecific
        Case 0:                     PlotNames.X = "PelvicTilt":    PlotNames.Y = "PelvicObl":      PlotNames.z = "PelvicRot"
        Case 1:                     PlotNames.X = "TrunkTilt":     PlotNames.Y = "TrunkObl":       PlotNames.z = "TrunkRot"
        End Select
    Case "HipJointAngle":           PlotNames.X = " Hip FE":       PlotNames.Y = " Hip AbdAdd":    PlotNames.z = " Hip IRER"
    Case "KneeJointAngle":          PlotNames.X = " Knee FE":      PlotNames.Y = " Knee VarVal":   PlotNames.z = " Knee IRER"
    Case "AnkleJointAngle":         PlotNames.X = " Ankle PFDF":   PlotNames.Y = "":               PlotNames.z = " Ankle IRER"
    Case "FootProgressionAngle"
        Select Case sSideLtr
        Case "L":                   PlotNames.X = "":              PlotNames.Y = "":               PlotNames.z = "eft" ' L put on below.
        Case "R":                   PlotNames.X = "":              PlotNames.Y = "":               PlotNames.z = "ight" ' R put on below.
        End Select
    '
    Case "HipJointMoment"
        Select Case sSideLtr
        Case "L":            PlotNames.X = " Hip Joint FE Moment":     PlotNames.Y = " Hip AbdAdd Joint Moment":    PlotNames.z = " Hip IRER Joint Moment"
        Case "R":            PlotNames.X = " Hip FE Moment":           PlotNames.Y = " Hip AbdAdd Joint Moment":    PlotNames.z = " Hip IRER Joint Moment"
        End Select
    Case "KneeJointMoment":  PlotNames.X = " Knee FE Joint Moment":    PlotNames.Y = " Knee VarVal Joint Moment":   PlotNames.z = " Knee IRER Joint Moment"
    Case "AnkleJointMoment": PlotNames.X = " Ankle PFDF Joint Moment": PlotNames.Y = " Ankle InvEver Joint Moment": PlotNames.z = " Ankle IRER Joint Moment"
    '
    Case "HipJointPower":    PlotNames.X = "":                         PlotNames.Y = "":                            PlotNames.z = " Hip Sag Joint Power"
    Case "KneeJointPower":   PlotNames.X = "":                         PlotNames.Y = "":                            PlotNames.z = " Knee Sag Joint Power"
    Case "AnkleJointPower":  PlotNames.X = "":                         PlotNames.Y = "":                            PlotNames.z = " Ankle Sag Joint Power"
    End Select
    '
    If Len(PlotNames.X) <> 0 Then PlotNames.X = sSideLtr + PlotNames.X
    If Len(PlotNames.Y) <> 0 Then PlotNames.Y = sSideLtr + PlotNames.Y
    If Len(PlotNames.z) <> 0 Then PlotNames.z = sSideLtr + PlotNames.z
    '
    DtaGetGraphPlotNames = PlotNames
End Function

Private Function DtaGetPointData(sSideLtr As String, sGraphTab As String, Optional iMoreSpecific As Long) As C3dPointFrameDataType()
    ' This returns all the frames for a particular point label.
    ' It's actually either kinematics or kinetics.
    '
    'Public Type C3dPointFrameDataType
    '    X As Single
    '    Y As Single
    '    z As Single
    '    Time As Single
    '    SeekPointerToNextPointFrameData As Long ' This one does not go to zero when there are no more records.  The caller must know how many frames to get.
    '    SeekPointerToNextPointLabelData As Long ' This one does not go to zero when there are no more records.  The caller must know how many points to get.
    '    StrideEventType As Long                 ' 0 = no event, 1 = Foot Strike, 2 = Foot Off.  Used for strides that make sense from strike/off.
    '    StrideEventSide As Long                 ' 1 = Left, 2 = Right.  Only used if StrideEventType <> OtherEvent.
    'End Type
    '
    Dim C3dPointData() As C3dPointFrameDataType
    Dim DtaPlotNames As DtaPlotNamesType
    Dim xCol As Long
    Dim yCol As Long
    Dim zCol As Long
    Dim iFrame As Long
    Dim iCol As Long
    Dim s As String
    '
    DtaPlotNames = DtaGetGraphPlotNames(sSideLtr, sGraphTab, iMoreSpecific)
    '
    xCol = DtaGraphPlotNameColumn(sGraphTab, DtaPlotNames.X)
    yCol = DtaGraphPlotNameColumn(sGraphTab, DtaPlotNames.Y)
    zCol = DtaGraphPlotNameColumn(sGraphTab, DtaPlotNames.z)
    '
    ReDim C3dPointData(1 To C3dHeader.PointFrameCount)
    '
    If xCol > 0 Then
        For iFrame = 1 To C3dHeader.PointFrameCount
            DtaGotoFrameLine iFrame
            For iCol = 1 To xCol
                s = DtaGetNextField
            Next iCol
            C3dPointData(iFrame).X = Val(s) '  * C3dHeader.ScaleFactor
        Next iFrame
    End If
    '
    If yCol > 0 Then
        For iFrame = 1 To C3dHeader.PointFrameCount
            DtaGotoFrameLine iFrame
            For iCol = 1 To yCol
                s = DtaGetNextField
            Next iCol
            C3dPointData(iFrame).Y = Val(s) '  * C3dHeader.ScaleFactor
        Next iFrame
    End If
    '
    If zCol > 0 Then
        For iFrame = 1 To C3dHeader.PointFrameCount
            DtaGotoFrameLine iFrame
            For iCol = 1 To zCol
                s = DtaGetNextField
            Next iCol
            C3dPointData(iFrame).z = Val(s) '  * C3dHeader.ScaleFactor
        Next iFrame
    End If
    '
    For iFrame = 1 To C3dHeader.PointFrameCount
        C3dPointData(iFrame).Time = (iFrame - 1) * (1 / C3dHeader.PointFrameRate) + C3dHeader.WalkStartTime
    Next iFrame
    '
    DtaGetPointData = C3dPointData
End Function

Private Function C3DLabelToDtaGraphTabXref(sLabel As String) As String
    ' This returns a Motion Analysis GraphTab name if it's not Vicon.
    ' This shouldn't be called unless giTypeOfMotionSystem=1, but a double-check is performed anyway.
    Dim s As String
    '
    If giTypeOfMotionSystem = 0 Then
        C3DLabelToDtaGraphTabXref = sLabel
    Else
        ' If it's the toe, we still go to the C3D file.
        Select Case sLabel
        Case "LTOE":                    C3DLabelToDtaGraphTabXref = "L.Toe"
        Case "RTOE":                    C3DLabelToDtaGraphTabXref = "R.Toe"
        Case Else
            '
            ' If it's any other label, we have the problem that the Motion Analysis labels aren't indicated with a side.
            ' Therefore, we just strip the side and return the overall label.  Getting the sides and planes (sagittal, transverse, axial)
            ' will be handled elsewhere.  We're also case sensitive.
            '
            s = Mid$(sLabel, 2)
            Select Case s
            Case "PelvisAngles":        C3DLabelToDtaGraphTabXref = "Trunk/Pelvis"
            Case "TrunkAbsAngles":      C3DLabelToDtaGraphTabXref = "Trunk/Pelvis"
            Case "HipAngles":           C3DLabelToDtaGraphTabXref = "HipJointAngle"
            Case "KneeAngles":          C3DLabelToDtaGraphTabXref = "KneeJointAngle"
            Case "AnkleAngles":         C3DLabelToDtaGraphTabXref = "AnkleJointAngle"
            Case "FootProgressAngles":  C3DLabelToDtaGraphTabXref = "FootProgressionAngle"
            '
            Case "HipMoment":           C3DLabelToDtaGraphTabXref = "HipJointMoment"
            Case "KneeMoment":          C3DLabelToDtaGraphTabXref = "KneeJointMoment"
            Case "AnkleMoment":         C3DLabelToDtaGraphTabXref = "AnkleJointMoment"
            '
            Case "HipPower":            C3DLabelToDtaGraphTabXref = "HipJointPower"
            Case "KneePower":           C3DLabelToDtaGraphTabXref = "KneeJointPower"
            Case "AnklePower":          C3DLabelToDtaGraphTabXref = "AnkleJointPower"
            End Select
        End Select
    End If
    ' if nothing was matched, LabelXRef just stays an empty string.
End Function

Private Sub DtaFileSetup()
    ' This should be called immediately after opening the file.
    Dim ptr As Long
    Dim s As String
    Dim i As Long
    Dim iLinePtr As Long
    Const MaxColumns = 500
    '
    DtaFrameNumber = 0
    DtaFileLine = ""
    DtaFileLinePtr = 0
    '
    ReDim DtaGraphTabName(1 To MaxColumns)
    DtaGotoBeginning
    Do While DtaFileLine <> "[Data]"
        DtaGetNextLine
        If DtaEof Then
            ReDim DtaGraphTabName(0)
            Exit Sub
        End If
    Loop
    '
    ' Now get the joint labels.
    DtaGetNextLine ' The very next line will have the joint names on it.
    ptr = 0
    For i = 1 To MaxColumns ' Just do plenty.
        s = DtaGetNextField
        If Len(s) <> 0 Then
            ptr = ptr + 1
            DtaGraphTabName(ptr).TheLabel = s
            DtaGraphTabName(ptr).TheColumn = i
        End If
    Next i
    '
    If ptr = 0 Then
        ReDim DtaGraphTabName(0)
    Else
        ReDim Preserve DtaGraphTabName(1 To ptr)
    End If
    '
    ' Now get the joint/plane labels.  In many cases, there are three for each joint.
    DtaGetNextLine ' The very next line will have the joint/plane names on it.
    If UBound(DtaGraphTabName) > 0 Then
        ReDim DtaGraphPlotName(1 To DtaGraphTabName(UBound(DtaGraphTabName)).TheColumn + 12)
        For i = 1 To DtaGraphTabName(UBound(DtaGraphTabName)).TheColumn + 12 ' Just do plenty past the last joint label.
            DtaGraphPlotName(i) = DtaGetNextField
        Next i
    Else
        ReDim DtaGraphPlotName(0)
    End If
    '
    ' Now we must make sure we can find the first frame line.
    ' It should be two more lines down.  We certainly shouldn't hit a blank line.
    Do
        DtaGetNextLine
        If DtaEof Then Exit Sub
        s = DtaGetNextField
        If Val(s) = 1 Then
            DtaFirstFrameLine = DtaFileLinePtr
            Exit Do
        End If
    Loop
    '
    ' Any additional setup here.
    '
End Sub

Private Function DtaGraphTabNameColumn(sGraphTabName As String) As Long
    ' Returns 0 if not found.  Case-sensitive.  Can be used to test existence.
    Dim i As Long
    '
    For i = 1 To UBound(DtaGraphTabName)
        If DtaGraphTabName(i).TheLabel = sGraphTabName Then
            DtaGraphTabNameColumn = DtaGraphTabName(i).TheColumn
            Exit Function
        End If
    Next i
    ' Just falls out if it's not found, or if DtaGraphTabName is dimmed to 0.
End Function

Public Function DtaGraphPlotNameExists(sGraphTabName As String, sGraphPlotName As String)
    DtaGraphPlotNameExists = DtaGraphPlotNameColumn(sGraphTabName, sGraphPlotName) <> 0
End Function


Private Function DtaGraphPlotNameColumn(sGraphTabName As String, sGraphPlotName As String) As Long
    ' Returns 0 if not found.  Case-sensitive.  Can be used to test existence.
    Dim i As Long
    Dim iTab As Long
    '
    iTab = DtaGraphTabNameColumn(sGraphTabName)
    If iTab > 0 Then
        For i = iTab To UBound(DtaGraphPlotName)
            If DtaGraphPlotName(i) = sGraphPlotName Then
                DtaGraphPlotNameColumn = i
                Exit Function
            End If
        Next i
    End If
    ' It just falls out if not found, or if DtaGraphPlotName is dimmed to 0.
End Function

Private Sub DtaGotoBeginning()
    Seek DtaFileNumber, 1
    DtaFrameNumber = 0
    DtaFileLine = ""
    DtaFileLinePtr = 0
End Sub

Private Sub DtaGotoFrameLine(iFrame As Long) ' iFrame should always be >=1.
    Dim i As Long
    '
    If DtaFileLinePtr >= ((DtaFirstFrameLine + iFrame) - 1) Then ' We must re-get to make sure we start at beginning of line.
        DtaGotoBeginning
        For i = 1 To ((DtaFirstFrameLine + iFrame) - 1)
            DtaGetNextLine
        Next i
    Else ' If we're behind it, save some time.
        For i = (DtaFileLinePtr + 1) To ((DtaFirstFrameLine + iFrame) - 1)
            DtaGetNextLine
        Next i
    End If
End Sub

Private Sub DtaGotoFirstFrameLine()
    Dim i As Long
    '
    DtaGotoBeginning
    For i = 1 To DtaFirstFrameLine
        DtaGetNextLine
    Next i
End Sub

Private Sub DtaGetNextLine()
    ' It just returns blank lines after the end of the file.
    If DtaEof Then
        DtaFileLine = ""
    Else
        Line Input #DtaFileNumber, DtaFileLine
    End If
    DtaFileLinePtr = DtaFileLinePtr + 1
End Sub

Private Function DtaEof()
    DtaEof = EOF(DtaFileNumber)
End Function

Private Function DtaGetNextField() As String
    ' This will just return empty strings if you read past the end of the line.
    If InStr(DtaFileLine, vbTab) <> 0 Then
        DtaGetNextField = Left$(DtaFileLine, InStr(DtaFileLine, vbTab) - 1)
        DtaFileLine = Mid$(DtaFileLine, InStr(DtaFileLine, vbTab) + 1)
    Else
        DtaGetNextField = DtaFileLine
        DtaFileLine = ""
    End If
End Function

Public Function OpenC3dFile(FileSpec As String) As Long
    Dim Subjects As C3dGroupAndParamType
    Dim ScaleParam As C3dGroupAndParamType
    Dim Labels As C3dGroupAndParamType
    Dim Descriptions As C3dGroupAndParamType
    Dim Prefix As C3dGroupAndParamType
    Dim FrameRate As C3dGroupAndParamType
    Dim C3dParamHeader As C3dParamHeaderType
    Dim SubjectCount As Long
    Dim ScaleFactor As Single
    Dim ptr As Long
    Dim sEvtFile As String
    Dim sDtaFile As String
    '
    ' FileSpec must be a fully specified file specification.
    '
    ' Returns any file open error.
    ' 0 = no error.
    ' 1 = disk i/o error while opening.
    ' 2 = not a valid C3D file.
    ' 3 = data not stored for Intel (IEEE) or DEC processor.
    ' 4 = more than one subject (patient) found in file.
    ' 5 = scale factor is corrupted.
    ' 6 = there were no point labels defined.
    ' 7 = there were no frames in the file.
    ' 8 = the point frame rate is zero.
    ' 9 = the EVT file can't be found.
    '
    If giTypeOfMotionSystem = 1 Then
        sEvtFile = Left$(FileSpec, Len(FileSpec) - 4) & ".evt"
        sDtaFile = Left$(FileSpec, Len(FileSpec) - 4) & ".data"
        If (Not bFileExists(sEvtFile)) Or (Not bFileExists(sDtaFile)) Then
            CloseC3dFile
            OpenC3dFile = 9
            Exit Function
        End If
    End If
    '
    CloseC3dFile
    C3dFileSpec = FileSpec
    EvtFileSpec = sEvtFile
    DtaFileSpec = sDtaFile
    On Error Resume Next
    C3dFileNumber = FreeFile
    Open C3dFileSpec For Binary As C3dFileNumber
    If Err Then
        CloseC3dFile
        OpenC3dFile = 1
        Exit Function
    End If
    If giTypeOfMotionSystem = 1 Then
        DtaFileNumber = FreeFile
        Open DtaFileSpec For Input As DtaFileNumber
        If Err Then
            CloseC3dFile
            OpenC3dFile = 1
            Exit Function
        End If
        DtaFileSetup
    End If
    On Error GoTo 0
    '
    ' Get the header.
    Get C3dFileNumber, 1, C3dHeaderRw   ' This one is used only when there is an intent to write to the file.
    Get C3dFileNumber, 1, C3dHeader
    If C3dHeader.FileID <> 80 Then
        CloseC3dFile
        OpenC3dFile = 2
        Exit Function
    End If
    '
    ' Get the param header.
    Seek C3dFileNumber, (C3dHeader.ParamBlockPointer - 1) * C3dBlockSize + 1
    Get C3dFileNumber, , C3dParamHeader
    If C3dParamHeader.ProcessorType <> 84 And C3dParamHeader.ProcessorType <> 85 Then
        CloseC3dFile
        OpenC3dFile = 3
        Exit Function
    End If
    C3dHeader.ProcessorType = C3dParamHeader.ProcessorType
    ' Set seek position immediately after param header is gotten.
    C3dHeader.SeekPtrForFirstC3dParam = Seek(C3dFileNumber)
    '
    ' Fix singles in header.
    FixSingle C3dHeader.ScaleFactor
    FixSingle C3dHeader.PointFrameRate
    FixSingleArray C3dHeader.EventTimes()
    '
    ' Some derived header variables.
    If C3dHeader.AnalogSamplesPerChannel > 0 Then
        C3dHeader.AnalogChannelsCount = C3dHeader.AnalogValuesPerFrame / C3dHeader.AnalogSamplesPerChannel
    Else
        C3dHeader.AnalogChannelsCount = 0
    End If
    C3dHeader.AnalogFrameRate = C3dHeader.PointFrameRate * C3dHeader.AnalogSamplesPerChannel
    C3dHeader.SeekPtrForFirstFrameData = (LongFromUnsignedInt(C3dHeader.FirstDataBlockPointer) - 1) * C3dBlockSize + 1
    C3dHeader.PointFrameCount = C3dHeader.LastPointFrameNumber - C3dHeader.FirstPointFrameNumber + 1
    '
    If C3dHeader.PointFrameCount < 1 Then
        CloseC3dFile
        OpenC3dFile = 7
        Exit Function
    End If
    '
    ' Set the frame rate.
    FrameRate = GetMatchingC3dGroupAndParam("TRIAL", "CAMERA_RATE")
    ' The "parameters" frame rate should take precedence, but if it doesn't exist, we'll take the one in the header.
    On Error Resume Next
    C3dHeader.PointFrameRate = FrameRate.C3dParam.Singles(1, 1, 1, 1, 1, 1, 1)
    On Error GoTo 0
    If C3dHeader.PointFrameRate < 1 Then
        FrameRate = GetMatchingC3dGroupAndParam("TRIAL", "CAMERA_RATE")
        CloseC3dFile
        OpenC3dFile = 8
        Exit Function
    End If
    '
    ' Make sure there's only one subject in the file.
    Subjects = GetMatchingC3dGroupAndParam("SUBJECTS", "USED")
    If Subjects.C3dGroup.ID <> 0 And Subjects.C3dParam.ID <> 0 Then
        On Error Resume Next
        SubjectCount = Subjects.C3dParam.Integers(1, 1, 1, 1, 1, 1, 1)
        On Error GoTo 0
        If SubjectCount > 1 Then ' It might be ZERO or ONE, and that's ok.
            CloseC3dFile
            OpenC3dFile = 4
            Exit Function
        End If
    End If
    '
    ' Make sure the parameter scale factor matches the header scale factor.
    ScaleParam = GetMatchingC3dGroupAndParam("POINT", "SCALE")
    If ScaleParam.C3dGroup.ID = 0 Or ScaleParam.C3dParam.ID = 0 Then
        CloseC3dFile
        OpenC3dFile = 5
        Exit Function
    Else
        On Error Resume Next
        ScaleFactor = ScaleParam.C3dParam.Singles(1, 1, 1, 1, 1, 1, 1)
        On Error GoTo 0
        If Err <> 0 Or ScaleFactor <> C3dHeader.ScaleFactor Or ScaleFactor = 0 Then
            CloseC3dFile
            OpenC3dFile = 5
            Exit Function
        End If
    End If
    ' Patch up scale factor stuff.
    If C3dHeader.ScaleFactor < 0 Then
        C3dHeader.DataStoredAsIntegerOrSingle = 2
    Else
        C3dHeader.DataStoredAsIntegerOrSingle = 1
    End If
    C3dHeader.ScaleFactor = Abs(C3dHeader.ScaleFactor)
    '
    ' Set prefix variables.
    Prefix = GetMatchingC3dGroupAndParam("SUBJECTS", "USES_PREFIXES")
    If Prefix.C3dGroup.ID <> 0 And Prefix.C3dParam.ID <> 0 Then
        On Error Resume Next
        C3dHeader.PrefixIsUsed = Prefix.C3dParam.Integers(1, 1, 1, 1, 1, 1, 1) <> 0
        On Error GoTo 0
    End If
    If C3dHeader.PrefixIsUsed Then
        Prefix = GetMatchingC3dGroupAndParam("SUBJECTS", "LABEL_PREFIXES")
        If Prefix.C3dGroup.ID <> 0 And Prefix.C3dParam.ID <> 0 Then
            On Error Resume Next
            C3dHeader.PrefixString = Prefix.C3dParam.Text(1, 1, 1, 1, 1, 1)
            On Error GoTo 0
        End If
    End If
    '
    ' Get the point labels for later use.
    If C3dHeader.PointLabelCount < 1 Then
        CloseC3dFile
        OpenC3dFile = 6
        Exit Function
    End If
    Erase C3dPointLabels
    ReDim C3dPointLabels(1 To C3dHeader.PointLabelCount)
    Labels = GetMatchingC3dGroupAndParam("POINT", "LABELS")
    If Labels.C3dGroup.ID <> 0 And Labels.C3dParam.ID <> 0 Then
        On Error Resume Next ' Ignore errors here.
        For ptr = 1 To C3dHeader.PointLabelCount
            C3dPointLabels(ptr) = Trim$(Labels.C3dParam.Text(ptr, 1, 1, 1, 1, 1))
            If Left$(UCase$(C3dPointLabels(ptr)), Len(C3dHeader.PrefixString)) = UCase$(C3dHeader.PrefixString) Then
                C3dPointLabels(ptr) = Mid$(C3dPointLabels(ptr), Len(C3dHeader.PrefixString) + 1)
            End If
        Next ptr
        On Error GoTo 0
    End If
    '
    ' Get the analog labels and descriptions for later use.
    Erase C3DAnalogLabels
    Erase C3DAnalogDescriptions
    If C3dHeader.AnalogChannelsCount <> 0 Then
        ReDim C3DAnalogLabels(1 To C3dHeader.AnalogChannelsCount)
        ReDim C3DAnalogDescriptions(1 To C3dHeader.AnalogChannelsCount)
        Labels = GetMatchingC3dGroupAndParam("ANALOG", "LABELS")
        Descriptions = GetMatchingC3dGroupAndParam("ANALOG", "DESCRIPTIONS")
        If Labels.C3dGroup.ID <> 0 And Labels.C3dParam.ID <> 0 And Descriptions.C3dGroup.ID <> 0 And Descriptions.C3dParam.ID <> 0 Then
            On Error Resume Next ' Ignore errors here.
            For ptr = 1 To C3dHeader.AnalogChannelsCount
                C3DAnalogLabels(ptr) = Trim$(Labels.C3dParam.Text(ptr, 1, 1, 1, 1, 1))
                If Left$(UCase$(C3DAnalogLabels(ptr)), Len(C3dHeader.PrefixString)) = UCase$(C3dHeader.PrefixString) Then
                    C3DAnalogLabels(ptr) = Mid$(C3DAnalogLabels(ptr), Len(C3dHeader.PrefixString) + 1)
                End If
                C3DAnalogDescriptions(ptr) = Trim$(Descriptions.C3dParam.Text(ptr, 1, 1, 1, 1, 1))
                If Left$(UCase$(C3DAnalogDescriptions(ptr)), Len(C3dHeader.PrefixString)) = UCase$(C3dHeader.PrefixString) Then
                    C3DAnalogDescriptions(ptr) = Mid$(C3DAnalogDescriptions(ptr), Len(C3dHeader.PrefixString) + 1)
                End If
            Next ptr
            On Error GoTo 0
        End If
    End If
    '
    ' Patch up some more header stuff.
    C3dHeader.ElapsedTimeOfWalk = (1 / C3dHeader.PointFrameRate) * (C3dHeader.PointFrameCount - 1)
    C3dHeader.TimeBetweenFrames = 1 / C3dHeader.PointFrameRate
    C3dHeader.WalkStartTime = (C3dHeader.FirstPointFrameNumber - 1) * C3dHeader.TimeBetweenFrames
    C3dHeader.WalkEndTime = (C3dHeader.LastPointFrameNumber - 1) * C3dHeader.TimeBetweenFrames
    '
    ' See if there are strike/off events.  If there are, set up an array for quick future identification to frames.
    SetStrikeOffEvents
    PrepareStrikeOffEventFramesArray
    '
    ' If we got to here, everything's ok.
    OpenC3dFile = 0
End Function

Public Sub CloseC3dFile()
    On Error Resume Next
    Close C3dFileNumber
    C3dFileNumber = 0
    Close DtaFileNumber
    DtaFileNumber = 0
    Erase DtaGraphTabName
End Sub

Public Function GetC3dPointDataFromLabelAndTagStrideEvents(Label As String) As C3dPointFrameDataType()
    Dim C3dPointData() As C3dPointFrameDataType
    Dim ptr As Long
    '
    ' If the label doesn't exist, we're done.
    If Not C3dPointLabelExists(Label) Then Exit Function
    '
    ' Get point data for label.
    C3dPointData = GetC3dPointDataFromLabel(Label)
    '
    ' If there aren't any strike/off events, we're done.
    If StrikeOffEventFramesCount < 1 Then
        GetC3dPointDataFromLabelAndTagStrideEvents = C3dPointData
        Exit Function
    End If
    '
    ' We've got everything, so mark the strike/off event frames.
    For ptr = 1 To StrikeOffEventFramesCount
        C3dPointData(StrikeOffEventFrames(ptr).FrameNumber).StrideEventSide = StrikeOffEventFrames(ptr).StrideEventSide
        C3dPointData(StrikeOffEventFrames(ptr).FrameNumber).StrideEventType = StrikeOffEventFrames(ptr).StrideEventType
    Next ptr
    '
    GetC3dPointDataFromLabelAndTagStrideEvents = C3dPointData
End Function

Public Function GetC3dAnalogDataFromLabelAndTagStrideEvents(Label As String) As C3dAnalogFrameDataType()
    Dim C3dAnalogData() As C3dAnalogFrameDataType
    Dim ptr As Long
    '
    ' If the label doesn't exist, we're done.
    If Not C3dAnalogLabelExists(Label) Then Exit Function
    '
    ' Get analog data for label.
    C3dAnalogData = GetC3dAnalogDataFromLabel(Label)
    '
    ' If there aren't any strike/off events, we're done.
    If StrikeOffEventFramesCount < 1 Then
        GetC3dAnalogDataFromLabelAndTagStrideEvents = C3dAnalogData
        Exit Function
    End If
    '
    ' We've got everything, so mark the strike/off event frames.
    For ptr = 1 To StrikeOffEventFramesCount
        C3dAnalogData(StrikeOffEventFrames(ptr).FrameNumber).StrideEventSide = StrikeOffEventFrames(ptr).StrideEventSide
        C3dAnalogData(StrikeOffEventFrames(ptr).FrameNumber).StrideEventType = StrikeOffEventFrames(ptr).StrideEventType
    Next ptr
    '
    GetC3dAnalogDataFromLabelAndTagStrideEvents = C3dAnalogData
End Function

Public Function GetC3dParamHeader() As C3dParamHeaderType
    ' This shouldn't be needed for normal processing.
    ' All necessary information has been stored in the main header record.
    Dim C3dParamHeader As C3dParamHeaderType
    '
    Seek C3dFileNumber, (C3dHeader.ParamBlockPointer - 1) * C3dBlockSize + 1
    Get C3dFileNumber, , C3dParamHeader
    GetC3dParamHeader = C3dParamHeader
End Function

Public Function C3dGroupExists(GroupName As String) As Boolean
    Dim C3dGroup As C3dParamType
    Dim SeekPtr As Long
    '
    SeekPtr = C3dHeader.SeekPtrForFirstC3dParam
    Do
        If SeekPtr = 0 Then Exit Do
        C3dGroup = GetC3dParam(SeekPtr)
        If C3dGroup.ID = 0 Then Exit Do
        If C3dGroup.IsGroupOrParam = 1 Then
            If C3dGroup.Name = GroupName Then
                C3dGroupExists = True
                Exit Do
            End If
        End If
        SeekPtr = C3dGroup.NextSeekPointer
    Loop
End Function

Public Function GetMatchingC3dGroupAndParam(GroupName As String, ParamName As String) As C3dGroupAndParamType
    ' The caller should check the C3dGroup.ID and C3dParam.ID to make sure things were found.
    ' If the group is not found, the param is irrelevant, but C3dParam.ID could still be checked.
    Dim C3dGroup As C3dParamType
    Dim C3dParam As C3dParamType
    Dim SeekPtr As Long
    Dim GroupFound As Boolean
    Dim ParamFound As Boolean
    Dim UcaseGroupName As String
    Dim UcaseParamName As String
    '
    UcaseGroupName = UCase$(Trim$(GroupName))
    UcaseParamName = UCase$(Trim$(ParamName))
    '
    SeekPtr = C3dHeader.SeekPtrForFirstC3dParam
    Do
        If SeekPtr = 0 Then Exit Do
        C3dGroup = GetC3dParam(SeekPtr)
        If C3dGroup.IsGroupOrParam = 1 And UCase$(C3dGroup.Name) = UcaseGroupName Then
            GroupFound = True
            Exit Do
        End If
        SeekPtr = C3dGroup.NextSeekPointer
    Loop
    If GroupFound Then
        SeekPtr = C3dHeader.SeekPtrForFirstC3dParam
        Do
            If SeekPtr = 0 Then Exit Do
            C3dParam = GetC3dParam(SeekPtr)
            If C3dParam.IsGroupOrParam = 2 And C3dParam.ID = C3dGroup.ID And UCase$(C3dParam.Name) = UcaseParamName Then
                ParamFound = True
                Exit Do
            End If
            SeekPtr = C3dParam.NextSeekPointer
        Loop
    End If
    If Not GroupFound Then C3dGroup.ID = 0
    If Not ParamFound Then C3dParam.ID = 0
    '
    GetMatchingC3dGroupAndParam.C3dGroup = C3dGroup
    GetMatchingC3dGroupAndParam.C3dParam = C3dParam
End Function

Public Function GetC3dParam(SeekPtr As Long, Optional bZeroLastSeekPointer As Boolean = True, _
                            Optional bLastGroupParamRecord As Boolean) As C3dParamType
    ' When starting, C3dHeader.SeekPtrForFirstC3dParam should be passed in.
    ' The bZeroLastSeekPointer and bLastGroupParamRecord are here only for routines
    ' that want to write additional group/param records.  The bLastGroupParamRecord
    ' is a return argument.
    Dim C3dParam As C3dParamType
    Dim NextC3dParam As C3dParamType
    Dim ptr As Long
    Dim p1 As Long
    Dim p2 As Long
    Dim p3 As Long
    Dim p4 As Long
    Dim p5 As Long
    Dim p6 As Long
    Dim p7 As Long
    Dim DataAvailable As Boolean
    '
    Get C3dFileNumber, SeekPtr, C3dParam.NameLength
    C3dParam.Locked = C3dParam.NameLength > 127
    NoSign C3dParam.NameLength
    Get C3dFileNumber, , C3dParam.ID
    If C3dParam.ID > 128 Then C3dParam.IsGroupOrParam = 1 Else C3dParam.IsGroupOrParam = 2
    NoSign C3dParam.ID
    If C3dParam.ID = 0 Then Exit Function
    C3dParam.Name = Space$(C3dParam.NameLength)
    Get C3dFileNumber, , C3dParam.Name: TrimIt C3dParam.Name
    Get C3dFileNumber, , C3dParam.OffsetForNext
    C3dParam.NextSeekPointer = Seek(C3dFileNumber) + C3dParam.OffsetForNext - 2
    If C3dParam.IsGroupOrParam = 2 Then ' Get extra stuff that's in param records (the data).
        Get C3dFileNumber, , C3dParam.DataType
        Get C3dFileNumber, , C3dParam.OriginalDimensionCount
        ' If the data is not dimensioned, we pretend it is anyway.
        ' In fact, we pad up to the possible max of MaxDimensions=7.
        If C3dParam.OriginalDimensionCount = 0 Then ' If this happens for text, it indicates one byte(character) of storage.
            ReDim C3dParam.DimValue(1 To C3dMaxDimensions)
            For ptr = 1 To C3dMaxDimensions
                C3dParam.DimValue(ptr) = 1
            Next ptr
        Else
            ReDim C3dParam.DimValue(1 To C3dMaxDimensions)
            For ptr = 1 To C3dParam.OriginalDimensionCount
                Get C3dFileNumber, , C3dParam.DimValue(ptr)
            Next ptr
            For ptr = C3dParam.OriginalDimensionCount + 1 To C3dMaxDimensions
                C3dParam.DimValue(ptr) = 1
            Next ptr
        End If
        '
        DataAvailable = True
        For ptr = 1 To C3dMaxDimensions
            If C3dParam.DimValue(ptr) = 0 Then
                C3dParam.DataType = 0 ' If there's no data, set the data type to ZERO.
                DataAvailable = False
                Exit For
            End If
        Next ptr
        '
        If DataAvailable Then
            Select Case C3dParam.DataType
            Case 255 ' (-1) Text.
                ' The first dimension is actually the string length, so we have to
                ' store this in another variable, and move down all the other dimensions.
                C3dParam.TextLength = C3dParam.DimValue(1)
                For ptr = 1 To C3dMaxDimensions - 1
                    C3dParam.DimValue(ptr) = C3dParam.DimValue(ptr + 1)
                Next ptr
                ReDim Preserve C3dParam.DimValue(1 To C3dMaxDimensions - 1)
                '
                ReDim C3dParam.Text(1 To C3dParam.DimValue(1), 1 To C3dParam.DimValue(2), 1 To C3dParam.DimValue(3), 1 To C3dParam.DimValue(4), 1 To C3dParam.DimValue(5), 1 To C3dParam.DimValue(6))
                For p6 = 1 To C3dParam.DimValue(6)
                For p5 = 1 To C3dParam.DimValue(5)
                For p4 = 1 To C3dParam.DimValue(4)
                For p3 = 1 To C3dParam.DimValue(3)
                For p2 = 1 To C3dParam.DimValue(2)
                For p1 = 1 To C3dParam.DimValue(1)
                    C3dParam.Text(p1, p2, p3, p4, p5, p6) = Space$(C3dParam.TextLength)
                    Get C3dFileNumber, , C3dParam.Text(p1, p2, p3, p4, p5, p6): TrimIt C3dParam.Text(p1, p2, p3, p4, p5, p6)
                Next p1
                Next p2
                Next p3
                Next p4
                Next p5
                Next p6
            Case 1 ' Byte.
                ReDim C3dParam.Bytes(1 To C3dParam.DimValue(1), 1 To C3dParam.DimValue(2), 1 To C3dParam.DimValue(3), 1 To C3dParam.DimValue(4), 1 To C3dParam.DimValue(5), 1 To C3dParam.DimValue(6), 1 To C3dParam.DimValue(7))
                For p7 = 1 To C3dParam.DimValue(7)
                For p6 = 1 To C3dParam.DimValue(6)
                For p5 = 1 To C3dParam.DimValue(5)
                For p4 = 1 To C3dParam.DimValue(4)
                For p3 = 1 To C3dParam.DimValue(3)
                For p2 = 1 To C3dParam.DimValue(2)
                For p1 = 1 To C3dParam.DimValue(1)
                    Get C3dFileNumber, , C3dParam.Bytes(p1, p2, p3, p4, p5, p6, p7)
                Next p1
                Next p2
                Next p3
                Next p4
                Next p5
                Next p6
                Next p7
            Case 2 ' Integer.
                ReDim C3dParam.Integers(1 To C3dParam.DimValue(1), 1 To C3dParam.DimValue(2), 1 To C3dParam.DimValue(3), 1 To C3dParam.DimValue(4), 1 To C3dParam.DimValue(5), 1 To C3dParam.DimValue(6), 1 To C3dParam.DimValue(7))
                For p7 = 1 To C3dParam.DimValue(7)
                For p6 = 1 To C3dParam.DimValue(6)
                For p5 = 1 To C3dParam.DimValue(5)
                For p4 = 1 To C3dParam.DimValue(4)
                For p3 = 1 To C3dParam.DimValue(3)
                For p2 = 1 To C3dParam.DimValue(2)
                For p1 = 1 To C3dParam.DimValue(1)
                    Get C3dFileNumber, , C3dParam.Integers(p1, p2, p3, p4, p5, p6, p7)
                Next p1
                Next p2
                Next p3
                Next p4
                Next p5
                Next p6
                Next p7
            Case 4 ' Single.
                ReDim C3dParam.Singles(1 To C3dParam.DimValue(1), 1 To C3dParam.DimValue(2), 1 To C3dParam.DimValue(3), 1 To C3dParam.DimValue(4), 1 To C3dParam.DimValue(5), 1 To C3dParam.DimValue(6), 1 To C3dParam.DimValue(7))
                For p7 = 1 To C3dParam.DimValue(7)
                For p6 = 1 To C3dParam.DimValue(6)
                For p5 = 1 To C3dParam.DimValue(5)
                For p4 = 1 To C3dParam.DimValue(4)
                For p3 = 1 To C3dParam.DimValue(3)
                For p2 = 1 To C3dParam.DimValue(2)
                For p1 = 1 To C3dParam.DimValue(1)
                    Get C3dFileNumber, , C3dParam.Singles(p1, p2, p3, p4, p5, p6, p7)
                    FixSingle C3dParam.Singles(p1, p2, p3, p4, p5, p6, p7)
                Next p1
                Next p2
                Next p3
                Next p4
                Next p5
                Next p6
                Next p7
            End Select
        End If
    End If
    '
    Get C3dFileNumber, , C3dParam.DescriptionLength
    NoSign C3dParam.DescriptionLength
    If C3dParam.DescriptionLength > 0 Then
        C3dParam.Description = Space$(C3dParam.DescriptionLength)
        Get C3dFileNumber, , C3dParam.Description: TrimIt C3dParam.Description
    End If
    '
    ' Now, the only thing left to do is to figure out if there are more param records
    ' left in the file.  We must peek at the next record to figure this out.
    bLastGroupParamRecord = False
    If C3dParam.OffsetForNext < 2 Then
        bLastGroupParamRecord = True
        If bZeroLastSeekPointer Then C3dParam.NextSeekPointer = 0
    Else
        Get C3dFileNumber, C3dParam.NextSeekPointer, NextC3dParam.NameLength
        Get C3dFileNumber, , NextC3dParam.ID
        NoSign NextC3dParam.ID
        If NextC3dParam.ID = 0 Then
            bLastGroupParamRecord = True
            If bZeroLastSeekPointer Then C3dParam.NextSeekPointer = 0
        End If
    End If
    '
    GetC3dParam = C3dParam
End Function

Public Function C3dPointLabelExists(Label As String) As Boolean
    If giTypeOfMotionSystem = 0 Then
        C3dPointLabelExists = GetC3DPointLabelPointer(Label) <> 0
    Else
        If Label = "LTOE" Or Label = "RTOE" Then ' We still go to the C3D file for the marker data.
            C3dPointLabelExists = GetC3DPointLabelPointer(C3DLabelToDtaGraphTabXref(Label)) <> 0
        Else
            C3dPointLabelExists = DtaGraphTabNameColumn(C3DLabelToDtaGraphTabXref(Label)) <> 0
        End If
    End If
End Function

Public Function GetC3DPointLabelPointer(Label As String) As Long
    ' Returns ZERO if not found
    Dim ptr As Long
    Dim s As String
    '
    s = UCase$(Trim$(Label))
    For ptr = 1 To C3dHeader.PointLabelCount
        If UCase$(C3dPointLabels(ptr)) = s Then
            GetC3DPointLabelPointer = ptr
            Exit Function
        End If
    Next ptr
    GetC3DPointLabelPointer = 0
End Function

Public Function C3dAnalogLabelExists(Label As String) As Boolean
    C3dAnalogLabelExists = GetC3DAnalogLabelPointer(Label) <> 0
End Function

Public Function GetC3DAnalogLabelPointer(Label As String) As Long
    ' Returns ZERO if not found
    Dim ptr As Long
    Dim s As String
    '
    s = UCase$(Trim$(Label))
    For ptr = 1 To C3dHeader.AnalogChannelsCount
        If UCase$(C3DAnalogLabels(ptr)) = s Then
            GetC3DAnalogLabelPointer = ptr
            Exit Function
        End If
    Next ptr
    GetC3DAnalogLabelPointer = 0
End Function

Public Function GetC3dPointDataFromLabel(Label As String) As C3dPointFrameDataType()
    ' Be careful with this procedure.  It will return an undimmed string if the label is not found.
    ' If the label is in C3dPointLabels(), it will be found.
    Dim ptr As Long
    Dim s As String
    Dim bGetFromC3D As Boolean
    Dim sGraphTab As String
    Dim sSideLtr As String
    Dim iMoreSpecific As Long
    '
    s = Label
    '
    ' Just do some quick cross-references here because only toe is used from C3D.
    If giTypeOfMotionSystem = 1 Then
        Select Case s
        Case "LTOE":                    s = "L.Toe": bGetFromC3D = True
        Case "RTOE":                    s = "R.Toe": bGetFromC3D = True
        Case Else
            sGraphTab = C3DLabelToDtaGraphTabXref(Label)
            If Mid$(Label, 2) = "TrunkAngles" Then iMoreSpecific = 1  ' This differentiates Pelvis from Trunk in the .DATA file.
            sSideLtr = Left$(Label, 1)
            If Len(sGraphTab) <> 0 Then
                GetC3dPointDataFromLabel = DtaGetPointData(sSideLtr, sGraphTab, iMoreSpecific)
            End If
            bGetFromC3D = False
        End Select
    Else
        bGetFromC3D = True
    End If
    '
    ' The toe still comes out of the C3D, even for Motion Analysis systems, so bUseDtaFile is not a pure Vicon flag.
    If bGetFromC3D Then
        s = UCase$(Trim$(s))
        For ptr = 1 To C3dHeader.PointLabelCount
            If UCase$(C3dPointLabels(ptr)) = s Then Exit For
        Next ptr
        If ptr <= C3dHeader.PointLabelCount Then
            GetC3dPointDataFromLabel = GetC3dPointDataFromNum(ptr)
        End If
    End If
End Function

Public Function GetC3dAnalogDataFromLabel(Label As String) As C3dAnalogFrameDataType()
    ' Be careful with this procedure.  It will return an undimmed string if the label is not found.
    ' If the label is in C3dAnalogLabels(), it will be found.
    Dim ptr As Long
    Dim s As String
    '
    s = UCase$(Trim$(Label))
    For ptr = 1 To C3dHeader.AnalogChannelsCount
        If UCase$(C3DAnalogLabels(ptr)) = s Then Exit For
    Next ptr
    If ptr <= C3dHeader.AnalogChannelsCount Then
        GetC3dAnalogDataFromLabel = GetC3dAnalogDataFromNum(ptr)
    End If
End Function

Public Function GetC3dPointDataFromNum(PointNumber As Long) As C3dPointFrameDataType()
    ' This returns all the frames for a particular point label.
    Dim C3dPointData() As C3dPointFrameDataType
    Dim ptr As Long
    Dim SeekPtr As Long
    '
    If PointNumber < 1 Or PointNumber > C3dHeader.PointLabelCount Then Exit Function
    ReDim C3dPointData(1 To C3dHeader.PointFrameCount)
    '
    ' Spin to correct point data record within the first frame.
    ' This could certainly be done easier by just figuring out the offset here, but whatever.
    SeekPtr = C3dHeader.SeekPtrForFirstFrameData
    For ptr = 1 To PointNumber
        C3dPointData(1) = GetC3dPointFrameData(SeekPtr)
        SeekPtr = C3dPointData(1).SeekPointerToNextPointLabelData ' Pointer to next label data, not frame data.
    Next ptr
    '
    ' Now, get the frames.  The first one is already in memory from above.
    SeekPtr = C3dPointData(1).SeekPointerToNextPointFrameData ' Pointer to next frame data, not label data.
    For ptr = 2 To C3dHeader.PointFrameCount
        C3dPointData(ptr) = GetC3dPointFrameData(SeekPtr)
        SeekPtr = C3dPointData(ptr).SeekPointerToNextPointFrameData
    Next ptr
    GetC3dPointDataFromNum = C3dPointData
End Function

Public Function GetC3dAnalogDataFromNum(ChannelNumber As Long) As C3dAnalogFrameDataType()
    ' This returns all the frames for a particular analog label.
    Dim C3dAnalogData() As C3dAnalogFrameDataType
    Dim C3dPointData As C3dPointFrameDataType
    Dim ptr As Long
    Dim SeekPtr As Long
    '
    If ChannelNumber < 1 Or ChannelNumber > C3dHeader.AnalogChannelsCount Then Exit Function
    ReDim C3dAnalogData(1 To C3dHeader.PointFrameCount) ' The frame count is always in terms of the 3D data, not the analog data.
    '
    ' First, set SeekPtr past all the 3D point data.
    ' Be careful, don't use same methods as points because the SeekPointerToNextPointLabelData gets set to zero when the end of the points is reached.
    SeekPtr = C3dHeader.SeekPtrForFirstFrameData
    Select Case C3dHeader.DataStoredAsIntegerOrSingle
    Case 1 ' Integer.
        SeekPtr = SeekPtr + C3dHeader.PointLabelCount * 8
    Case 2 ' Single.
        SeekPtr = SeekPtr + C3dHeader.PointLabelCount * 16
    End Select
    '
    ' Now, add an offset for the analog channel we want.  SeekPtr is still set from above to the first analog channel in the first 3D frame.
    Select Case C3dHeader.DataStoredAsIntegerOrSingle
    Case 1 ' Integer.
        SeekPtr = SeekPtr + (ChannelNumber - 1) * 2
    Case 2 ' Single.
        SeekPtr = SeekPtr + (ChannelNumber - 1) * 4
    End Select
    '
    ' Get first frame's data.
    C3dAnalogData(1) = GetC3dAnalogFrameData(SeekPtr)
    '
    ' Now, get the frames.  The first one is already in memory from above.
    SeekPtr = C3dAnalogData(1).SeekPointerToNextFrameData ' Pointer to next frame data, not label data.
    For ptr = 2 To C3dHeader.PointFrameCount ' Again, the frames are always according to the 3D data, not the analog data.
        C3dAnalogData(ptr) = GetC3dAnalogFrameData(SeekPtr)
        SeekPtr = C3dAnalogData(ptr).SeekPointerToNextFrameData
    Next ptr
    GetC3dAnalogDataFromNum = C3dAnalogData
End Function

Private Function GetC3dAnalogFrameData(SeekPtr As Long) As C3dAnalogFrameDataType
    ' At a minimum, the user should spin past the point (3D) data before calling this function.
    ' Ideally, the user will also have spun to the correct analog channel using repeated calls to this function.
    ' See the GetC3dPointFrameData function for more comments.
    '
    Dim C3dData As C3dAnalogFrameDataType
    Dim FrameRecordSize As Long
    Dim ChannelRecordSize As Long
    Dim FrameNumber As Long
    Dim ChannelNumber As Long
    Dim ptr As Long
    Dim Xint As Integer
    Dim Xsingle As Single
    '
    ' Dim the analog data variable for the analog timeslices per 3D frame.
    ReDim C3dData.X(1 To C3dHeader.AnalogSamplesPerChannel)
    '
    ' We figure out the next seek pointer first so we can mess with the value of SeekPointer.
    Select Case C3dHeader.DataStoredAsIntegerOrSingle
    Case 1 ' Integer.
        FrameRecordSize = C3dHeader.PointLabelCount * 8 + C3dHeader.AnalogValuesPerFrame * 2
    Case 2 ' Single.
        FrameRecordSize = C3dHeader.PointLabelCount * 16 + C3dHeader.AnalogValuesPerFrame * 4
    End Select
    C3dData.SeekPointerToNextFrameData = SeekPtr + FrameRecordSize
    '
    ' The C3dHeader.FrameCount variable must be counted to not overrun the file.
    ' There should be exactly C3dHeader.FrameCount frames in this file.
    ' However, SeekPointerToNextFrameData is set to ZERO when appropriate.
    '
    FrameNumber = ((SeekPtr - C3dHeader.SeekPtrForFirstFrameData) \ FrameRecordSize) + 1
    If FrameNumber = C3dHeader.PointFrameCount Then C3dData.SeekPointerToNextFrameData = 0
    '
    ' Calculate the time for the frame.
    C3dData.Time = (FrameNumber - 1) * (1 / C3dHeader.PointFrameRate) + C3dHeader.WalkStartTime
    '
    ' Now we can get the analog data for the 3D frame.
    Select Case C3dHeader.DataStoredAsIntegerOrSingle
    Case 1 ' Integer.
        For ptr = 1 To C3dHeader.AnalogSamplesPerChannel
            Get C3dFileNumber, SeekPtr, Xint
            C3dData.X(ptr) = Xint
            SeekPtr = SeekPtr + C3dHeader.AnalogChannelsCount * 2
        Next ptr
    Case 2 ' Single.
        For ptr = 1 To C3dHeader.AnalogSamplesPerChannel
            Get C3dFileNumber, SeekPtr, Xsingle
            FixSingle Xsingle
            C3dData.X(ptr) = Xsingle
            SeekPtr = SeekPtr + C3dHeader.AnalogChannelsCount * 4
        Next ptr
    End Select
    '
    GetC3dAnalogFrameData = C3dData
End Function

Private Function GetC3dPointFrameData(SeekPtr As Long) As C3dPointFrameDataType
    ' This is private because it really shouldn't be used outside of this module.
    '
    ' When starting, C3dHeader.SeekPtrForFirstFrameData should be passed in.
    '
    ' All the Point and Analog data for an entire frame will be sequentially stored,
    ' then you proceed to the next frame.  The frames will sequentially follow one another.
    '
    ' Frames are organized as follows:
    '   Point data records (there will be C3dHeader.PointLabelCount of these records).
    '   Analog sample and channel data.  The analog values are either two bytes or four bytes
    '     depending on the value of C3dHeader.DataStoredAsIntegerOrSingle.  There will always be
    '     exactly C3dHeader.AnalogValuesPerFrame of these analog values following each set
    '     of point data records.
    '
    Dim C3dIntegerData As C3dPointFrameIntegerDataType
    Dim C3dSingleData As C3dPointFrameSingleDataType
    Dim C3dData As C3dPointFrameDataType
    Dim FrameRecordSize As Long
    Dim LabelRecordSize As Long
    Dim FrameNumber As Long
    Dim LabelNumber As Long
    '
    Seek C3dFileNumber, SeekPtr
    Select Case C3dHeader.DataStoredAsIntegerOrSingle
    Case 1 ' Integer.
        Get C3dFileNumber, , C3dIntegerData
        C3dData.X = C3dIntegerData.X * C3dHeader.ScaleFactor
        C3dData.Y = C3dIntegerData.Y * C3dHeader.ScaleFactor
        C3dData.z = C3dIntegerData.z * C3dHeader.ScaleFactor
        FrameRecordSize = C3dHeader.PointLabelCount * 8 + C3dHeader.AnalogValuesPerFrame * 2
        LabelRecordSize = 8
    Case 2 ' Single.
        Get C3dFileNumber, , C3dSingleData
        FixSingle C3dSingleData.X
        FixSingle C3dSingleData.Y
        FixSingle C3dSingleData.z
        C3dData.X = C3dSingleData.X
        C3dData.Y = C3dSingleData.Y
        C3dData.z = C3dSingleData.z
        FrameRecordSize = C3dHeader.PointLabelCount * 16 + C3dHeader.AnalogValuesPerFrame * 4
        LabelRecordSize = 16
    End Select
    C3dData.SeekPointerToNextPointFrameData = SeekPtr + FrameRecordSize
    C3dData.SeekPointerToNextPointLabelData = Seek(C3dFileNumber) ' This should be used only on the first frame to get to the correct point.  After that, SeekPointerToNextPointFrameData can be used.
    '
    ' The C3dHeader.FrameCount variable must be counted to not overrun the file.
    ' There should be exactly C3dHeader.FrameCount frames in this file.
    ' However, SeekPointerToNextPointFrameData and SeekPointerToNextPointLabelData are set to ZERO when appropriate.
    '
    LabelNumber = ((SeekPtr - C3dHeader.SeekPtrForFirstFrameData) Mod FrameRecordSize) \ LabelRecordSize + 1
    FrameNumber = ((SeekPtr - C3dHeader.SeekPtrForFirstFrameData) \ FrameRecordSize) + 1
    '
    If LabelNumber = C3dHeader.PointLabelCount Then C3dData.SeekPointerToNextPointLabelData = 0
    If FrameNumber = C3dHeader.PointFrameCount Then C3dData.SeekPointerToNextPointFrameData = 0
    '
    ' Calculate the time for the frame.
    C3dData.Time = (FrameNumber - 1) * (1 / C3dHeader.PointFrameRate) + C3dHeader.WalkStartTime
    '
    GetC3dPointFrameData = C3dData
End Function

Private Sub FixSingle(n As Single)
    If C3dHeader.ProcessorType = 85 Then n = IeeeFromDec(n)
End Sub

Private Sub FixSingleArray(n() As Single)
    Dim i As Long
    '
    If C3dHeader.ProcessorType = 85 Then
        For i = LBound(n) To UBound(n)
            n(i) = IeeeFromDec(n(i))
        Next i
    End If
End Sub

Private Sub ReadTheEvtFileEventLines(EvtRtoLine As String, EvtRhsLine As String, EvtLtoLine As String, EvtLhsLine As String)
    Dim iFle As Long
    Dim sLine As String
    Dim Q As String
    '
    Q = Chr$(34)
    iFle = FreeFile
    Open EvtFileSpec For Input As iFle
    '
    Do While Not EOF(iFle)
        Line Input #iFle, sLine
        sLine = Trim$(UCase$(sLine))
        Select Case True
        Case Left$(sLine, 5) = Q & "RTO" & Q
            EvtRtoLine = sLine
        Case Left$(sLine, 5) = Q & "RHS" & Q
            EvtRhsLine = sLine
        Case Left$(sLine, 5) = Q & "LTO" & Q
            EvtLtoLine = sLine
        Case Left$(sLine, 5) = Q & "LHS" & Q
            EvtLhsLine = sLine
        End Select
    Loop
    Close iFle
End Sub

Private Function ParseEvtEventLine(ByVal TheLine As String) As Long()
    Dim iFrame() As Long
    Dim iPtr As Long
    Dim i As Long
    '
    ReDim iFrame(1 To 20)
    '
    ' The first part of the line is the type of event, which we'll ignore.
    TheLine = TheLine & " " ' Add a space to the end to make things easier.
    Do
        TheLine = Mid$(TheLine, InStr(TheLine, " ") + 1)
        If Len(TheLine) = 0 Then Exit Do
        iPtr = iPtr + 1
        iFrame(iPtr) = Val(Left$(TheLine, InStr(TheLine, " ") - 1)) + 1 ' The +1 is to account for the events being ZERO based in the EVT file.
    Loop
    '
    If iPtr > 0 Then
        ReDim Preserve iFrame(1 To iPtr)
    Else
        ReDim iFrame(0) ' This is just a signal that nothing was found.
    End If
    '
    ParseEvtEventLine = iFrame
End Function

Private Function GetTimeFromFrameAndCheck(iFrame As Long) As Single
    Dim TheTime As Single
    '
    TheTime = (iFrame - 1) * C3dHeader.TimeBetweenFrames
    If TheTime < 0 Then TheTime = 0
    If TheTime < C3dHeader.WalkStartTime Then TheTime = C3dHeader.WalkStartTime
    If TheTime > C3dHeader.WalkEndTime Then TheTime = C3dHeader.WalkEndTime
    '
    GetTimeFromFrameAndCheck = TheTime
End Function

Private Sub SetStrikeOffEvents()
    Dim Used As C3dGroupAndParamType
    Dim Contexts As C3dGroupAndParamType
    Dim Labels As C3dGroupAndParamType
    Dim Times As C3dGroupAndParamType
    Dim ptr As Long
    Dim SidedStrikeOffCount As Long
    Dim SidedGeneralCount As Long
    Dim NonSidedCount As Long
    Dim TheTime As Single
    ' These are used for finding valid strides.
    Dim cnt As Long
    Dim TempEvents() As StrikeOffSingleEventType
    Dim FoundOppositeOff As Boolean
    Dim FoundOppositeStrike As Boolean
    Dim FoundSameOff As Boolean
    Dim FoundSameStrike As Boolean
    Dim LeftStrideCount As Long
    Dim RightStrideCount As Long
    Dim DupeFound As Boolean
    '
    Dim EvtRtoLine As String
    Dim EvtRhsLine As String
    Dim EvtLtoLine As String
    Dim EvtLhsLine As String
    '
    Dim EvtRto() As Long
    Dim EvtRhs() As Long
    Dim EvtLto() As Long
    Dim EvtLhs() As Long
    '
    ' This assures it starts out blank.
    StrikeOffEvents = BlankStrikeOffEvents
    '
    Select Case giTypeOfMotionSystem
    Case 0 ' Vicon.
        Used = GetMatchingC3dGroupAndParam("EVENT", "USED")
        Contexts = GetMatchingC3dGroupAndParam("EVENT", "CONTEXTS")
        Labels = GetMatchingC3dGroupAndParam("EVENT", "LABELS")
        Times = GetMatchingC3dGroupAndParam("EVENT", "TIMES")
        '
        ' Error trapping in case the group or parameter doesn't exist, or if there was no data.
        On Error Resume Next
        StrikeOffEvents.TotalCount = Used.C3dParam.Integers(1, 1, 1, 1, 1, 1, 1)
        On Error GoTo 0
        '
        ' Make sure there are some events to get.
        If StrikeOffEvents.TotalCount = 0 Then Exit Sub
        '
        ' Count sub-event types.
        On Error GoTo ErrorInStrikeOffEvents
        For ptr = 1 To StrikeOffEvents.TotalCount
            Select Case UCase$(Contexts.C3dParam.Text(ptr, 1, 1, 1, 1, 1))
            Case "LEFT", "RIGHT"
                Select Case UCase$(Labels.C3dParam.Text(ptr, 1, 1, 1, 1, 1))
                Case "FOOT STRIKE", "FOOT OFF"
                    SidedStrikeOffCount = SidedStrikeOffCount + 1
                Case Else
                    SidedGeneralCount = SidedGeneralCount + 1
                End Select
            Case Else
                NonSidedCount = NonSidedCount + 1
            End Select
        Next ptr
    Case 1 ' Motion Analysis Corp. with EVT file.
        ReadTheEvtFileEventLines EvtRtoLine, EvtRhsLine, EvtLtoLine, EvtLhsLine
        '
        EvtRto = ParseEvtEventLine(EvtRtoLine)
        EvtRhs = ParseEvtEventLine(EvtRhsLine)
        EvtLto = ParseEvtEventLine(EvtLtoLine)
        EvtLhs = ParseEvtEventLine(EvtLhsLine)
        '
        StrikeOffEvents.TotalCount = UBound(EvtRto) + UBound(EvtRhs) + UBound(EvtLto) + UBound(EvtLhs)
        SidedStrikeOffCount = StrikeOffEvents.TotalCount
        '
        ' Make sure there are some events to get.
        If StrikeOffEvents.TotalCount = 0 Then Exit Sub
        If (UBound(EvtRto) < 1) Or (UBound(EvtRhs) < 1) Or (UBound(EvtLto) < 1) Or (UBound(EvtLhs) < 1) Then Exit Sub
    Case Else
        Exit Sub
    End Select
    '
    StrikeOffEvents.SidedStrikeOffCount = SidedStrikeOffCount
    StrikeOffEvents.SidedGeneralCount = SidedGeneralCount
    StrikeOffEvents.NonSidedCount = NonSidedCount
    '
    ' Dimension variables.  Dim to (0) if there is no data so that ubound can be used.
    If SidedStrikeOffCount > 0 Then ReDim StrikeOffEvents.SidedStrikeOffEvents(1 To SidedStrikeOffCount) Else ReDim StrikeOffEvents.SidedStrikeOffEvents(0)
    If SidedGeneralCount > 0 Then ReDim StrikeOffEvents.SidedGeneralEvents(1 To SidedGeneralCount) Else ReDim StrikeOffEvents.SidedGeneralEvents(0)
    If NonSidedCount > 0 Then ReDim StrikeOffEvents.NonSidedEvents(1 To NonSidedCount) Else ReDim StrikeOffEvents.NonSidedEvents(0)
    '
    ' The following will now be used as incremental pointers.
    SidedStrikeOffCount = 0
    SidedGeneralCount = 0
    NonSidedCount = 0
    '
    ' Now get the events.
    Select Case giTypeOfMotionSystem
    Case 0 ' Vicon.
        For ptr = 1 To StrikeOffEvents.TotalCount
            TheTime = Times.C3dParam.Singles(2, ptr, 1, 1, 1, 1, 1)
            If TheTime < 0 Then TheTime = 0
            If TheTime < C3dHeader.WalkStartTime Then TheTime = C3dHeader.WalkStartTime
            If TheTime > C3dHeader.WalkEndTime Then TheTime = C3dHeader.WalkEndTime
            '
            Select Case UCase$(Contexts.C3dParam.Text(ptr, 1, 1, 1, 1, 1))
            Case "LEFT"
                Select Case UCase$(Labels.C3dParam.Text(ptr, 1, 1, 1, 1, 1))
                Case "FOOT STRIKE"
                    SidedStrikeOffCount = SidedStrikeOffCount + 1
                    StrikeOffEvents.SidedStrikeOffEvents(SidedStrikeOffCount).Side = LeftSide
                    StrikeOffEvents.SidedStrikeOffEvents(SidedStrikeOffCount).Type = FootStrike
                    StrikeOffEvents.SidedStrikeOffEvents(SidedStrikeOffCount).Time = TheTime
                Case "FOOT OFF"
                    SidedStrikeOffCount = SidedStrikeOffCount + 1
                    StrikeOffEvents.SidedStrikeOffEvents(SidedStrikeOffCount).Side = LeftSide
                    StrikeOffEvents.SidedStrikeOffEvents(SidedStrikeOffCount).Type = FootOff
                    StrikeOffEvents.SidedStrikeOffEvents(SidedStrikeOffCount).Time = TheTime
                Case Else
                    SidedGeneralCount = SidedGeneralCount + 1
                    StrikeOffEvents.SidedGeneralEvents(SidedGeneralCount).Side = LeftSide
                    StrikeOffEvents.SidedGeneralEvents(SidedGeneralCount).Type = OtherEvent
                    StrikeOffEvents.SidedGeneralEvents(SidedGeneralCount).Time = TheTime
                End Select
            Case "RIGHT"
                Select Case UCase$(Labels.C3dParam.Text(ptr, 1, 1, 1, 1, 1))
                Case "FOOT STRIKE"
                    SidedStrikeOffCount = SidedStrikeOffCount + 1
                    StrikeOffEvents.SidedStrikeOffEvents(SidedStrikeOffCount).Side = RightSide
                    StrikeOffEvents.SidedStrikeOffEvents(SidedStrikeOffCount).Type = FootStrike
                    StrikeOffEvents.SidedStrikeOffEvents(SidedStrikeOffCount).Time = TheTime
                Case "FOOT OFF"
                    SidedStrikeOffCount = SidedStrikeOffCount + 1
                    StrikeOffEvents.SidedStrikeOffEvents(SidedStrikeOffCount).Side = RightSide
                    StrikeOffEvents.SidedStrikeOffEvents(SidedStrikeOffCount).Type = FootOff
                    StrikeOffEvents.SidedStrikeOffEvents(SidedStrikeOffCount).Time = TheTime
                Case Else
                    SidedGeneralCount = SidedGeneralCount + 1
                    StrikeOffEvents.SidedGeneralEvents(SidedGeneralCount).Side = RightSide
                    StrikeOffEvents.SidedGeneralEvents(SidedGeneralCount).Type = OtherEvent
                    StrikeOffEvents.SidedGeneralEvents(SidedGeneralCount).Time = TheTime
                End Select
            Case Else
                NonSidedCount = NonSidedCount + 1
                StrikeOffEvents.NonSidedEvents(NonSidedCount).Side = NoSide
                Select Case UCase$(Labels.C3dParam.Text(ptr, 1, 1, 1, 1, 1))
                Case "FOOT STRIKE"
                    StrikeOffEvents.NonSidedEvents(NonSidedCount).Type = FootStrike
                Case "FOOT OFF"
                    StrikeOffEvents.NonSidedEvents(NonSidedCount).Type = FootOff
                Case Else
                    StrikeOffEvents.NonSidedEvents(NonSidedCount).Type = OtherEvent
                End Select
                StrikeOffEvents.NonSidedEvents(NonSidedCount).Time = TheTime
            End Select
        Next ptr
    Case 1 ' Motion Analysis Corp. with EVT file.
        For ptr = 1 To UBound(EvtRto)
            TheTime = GetTimeFromFrameAndCheck(EvtRto(ptr))
            SidedStrikeOffCount = SidedStrikeOffCount + 1
            StrikeOffEvents.SidedStrikeOffEvents(SidedStrikeOffCount).Side = RightSide
            StrikeOffEvents.SidedStrikeOffEvents(SidedStrikeOffCount).Type = FootOff
            StrikeOffEvents.SidedStrikeOffEvents(SidedStrikeOffCount).Time = TheTime
        Next ptr
        For ptr = 1 To UBound(EvtRhs)
            TheTime = GetTimeFromFrameAndCheck(EvtRhs(ptr))
            SidedStrikeOffCount = SidedStrikeOffCount + 1
            StrikeOffEvents.SidedStrikeOffEvents(SidedStrikeOffCount).Side = RightSide
            StrikeOffEvents.SidedStrikeOffEvents(SidedStrikeOffCount).Type = FootStrike
            StrikeOffEvents.SidedStrikeOffEvents(SidedStrikeOffCount).Time = TheTime
        Next ptr
        For ptr = 1 To UBound(EvtLto)
            TheTime = GetTimeFromFrameAndCheck(EvtLto(ptr))
            SidedStrikeOffCount = SidedStrikeOffCount + 1
            StrikeOffEvents.SidedStrikeOffEvents(SidedStrikeOffCount).Side = LeftSide
            StrikeOffEvents.SidedStrikeOffEvents(SidedStrikeOffCount).Type = FootOff
            StrikeOffEvents.SidedStrikeOffEvents(SidedStrikeOffCount).Time = TheTime
        Next ptr
        For ptr = 1 To UBound(EvtLhs)
            TheTime = GetTimeFromFrameAndCheck(EvtLhs(ptr))
            SidedStrikeOffCount = SidedStrikeOffCount + 1
            StrikeOffEvents.SidedStrikeOffEvents(SidedStrikeOffCount).Side = LeftSide
            StrikeOffEvents.SidedStrikeOffEvents(SidedStrikeOffCount).Type = FootStrike
            StrikeOffEvents.SidedStrikeOffEvents(SidedStrikeOffCount).Time = TheTime
        Next ptr
    Case Else
        Exit Sub
    End Select
    '
    ' Sort the events into time order.
    SortStrikeOffEventGroup StrikeOffEvents.SidedStrikeOffCount, StrikeOffEvents.SidedStrikeOffEvents()
    SortStrikeOffEventGroup StrikeOffEvents.SidedGeneralCount, StrikeOffEvents.SidedGeneralEvents()
    SortStrikeOffEventGroup StrikeOffEvents.NonSidedCount, StrikeOffEvents.NonSidedEvents()
    '
    '
    ' Now we must find the valid stride events within the strike/off events.
    '
    '
    ' We will use a temp variable for convenience.
    TempEvents = StrikeOffEvents.SidedStrikeOffEvents
    '
    ' We will examine each foot strike to see if it sequences correctly: opposite-off, opposite-strike, same-off, same-strike.
    ' Also, there must be at least one frame difference between events.
    ' First, we must count them so we can dimension our array.
    For ptr = 1 To StrikeOffEvents.SidedStrikeOffCount - 4 ' We must have at least four events following to have a valid stride.
        If TempEvents(ptr).Type = FootStrike Then
            FoundOppositeOff = False
            If TempEvents(ptr + 1).Type = FootOff And TempEvents(ptr + 1).Side <> TempEvents(ptr).Side Then
                If TempEvents(ptr + 1).Time - TempEvents(ptr).Time > C3dHeader.TimeBetweenFrames * 0.8! Then
                    FoundOppositeOff = True
                End If
            End If
            FoundOppositeStrike = False
            If TempEvents(ptr + 2).Type = FootStrike And TempEvents(ptr + 2).Side <> TempEvents(ptr).Side Then
                If TempEvents(ptr + 2).Time - TempEvents(ptr + 1).Time > C3dHeader.TimeBetweenFrames * 0.8! Then
                    FoundOppositeStrike = True
                End If
            End If
            FoundSameOff = False
            If TempEvents(ptr + 3).Type = FootOff And TempEvents(ptr + 3).Side = TempEvents(ptr).Side Then
                If TempEvents(ptr + 3).Time - TempEvents(ptr + 2).Time > C3dHeader.TimeBetweenFrames * 0.8! Then
                    FoundSameOff = True
                End If
            End If
            FoundSameStrike = False
            If TempEvents(ptr + 4).Type = FootStrike And TempEvents(ptr + 4).Side = TempEvents(ptr).Side Then
                If TempEvents(ptr + 4).Time - TempEvents(ptr + 3).Time > C3dHeader.TimeBetweenFrames * 0.8! Then
                    FoundSameStrike = True
                End If
            End If
            If FoundOppositeOff And FoundOppositeStrike And FoundSameOff And FoundSameStrike Then
                StrikeOffEvents.ValidStrideCount = StrikeOffEvents.ValidStrideCount + 1
            End If
        End If
    Next ptr
    ' Now, we actually store them.
    If StrikeOffEvents.ValidStrideCount > 0 Then ' But make sure we have some to process.
        ReDim StrikeOffEvents.ValidStrides(1 To StrikeOffEvents.ValidStrideCount)
        cnt = 0
        For ptr = 1 To StrikeOffEvents.SidedStrikeOffCount - 4 ' We must have at least four events following to have a valid stride.
            If TempEvents(ptr).Type = FootStrike Then
                FoundOppositeOff = False
                If TempEvents(ptr + 1).Type = FootOff And TempEvents(ptr + 1).Side <> TempEvents(ptr).Side Then
                    If TempEvents(ptr + 1).Time - TempEvents(ptr).Time > C3dHeader.TimeBetweenFrames * 0.8! Then
                        FoundOppositeOff = True
                    End If
                End If
                FoundOppositeStrike = False
                If TempEvents(ptr + 2).Type = FootStrike And TempEvents(ptr + 2).Side <> TempEvents(ptr).Side Then
                    If TempEvents(ptr + 2).Time - TempEvents(ptr + 1).Time > C3dHeader.TimeBetweenFrames * 0.8! Then
                        FoundOppositeStrike = True
                    End If
                End If
                FoundSameOff = False
                If TempEvents(ptr + 3).Type = FootOff And TempEvents(ptr + 3).Side = TempEvents(ptr).Side Then
                    If TempEvents(ptr + 3).Time - TempEvents(ptr + 2).Time > C3dHeader.TimeBetweenFrames * 0.8! Then
                        FoundSameOff = True
                    End If
                End If
                FoundSameStrike = False
                If TempEvents(ptr + 4).Type = FootStrike And TempEvents(ptr + 4).Side = TempEvents(ptr).Side Then
                    If TempEvents(ptr + 4).Time - TempEvents(ptr + 3).Time > C3dHeader.TimeBetweenFrames * 0.8! Then
                        FoundSameStrike = True
                    End If
                End If
                If FoundOppositeOff And FoundOppositeStrike And FoundSameOff And FoundSameStrike Then
                    cnt = cnt + 1
                    StrikeOffEvents.ValidStrides(cnt).Side = TempEvents(ptr).Side
                    Select Case TempEvents(ptr).Side
                    Case LeftSide
                        LeftStrideCount = LeftStrideCount + 1
                        StrikeOffEvents.ValidStrides(cnt).NumberForSide = LeftStrideCount
                    Case RightSide
                        RightStrideCount = RightStrideCount + 1
                        StrikeOffEvents.ValidStrides(cnt).NumberForSide = RightStrideCount
                    End Select
                    ' Save the valid stride information.
                    StrikeOffEvents.ValidStrides(cnt).StartStrikeTime = TempEvents(ptr).Time
                    StrikeOffEvents.ValidStrides(cnt).OppositeFootOffTime = TempEvents(ptr + 1).Time
                    StrikeOffEvents.ValidStrides(cnt).OppositeFootStrikeTime = TempEvents(ptr + 2).Time
                    StrikeOffEvents.ValidStrides(cnt).SameFootOffTime = TempEvents(ptr + 3).Time
                    StrikeOffEvents.ValidStrides(cnt).StopStrikeTime = TempEvents(ptr + 4).Time
                End If
            End If
        Next ptr
        '
        ' Now we must stuff the valid stride events into the ValidStrikeOffEvents array so they can be matched with frames.
        StrikeOffEvents.ValidStrikeOffCount = StrikeOffEvents.ValidStrideCount * 5 ' This is, in all likelihood, inflated, but we will eliminate duplicates later.
        ReDim StrikeOffEvents.ValidStrikeOffEvents(1 To StrikeOffEvents.ValidStrikeOffCount)
        cnt = 0
        For ptr = 1 To StrikeOffEvents.ValidStrideCount
            cnt = cnt + 1
            StrikeOffEvents.ValidStrikeOffEvents(cnt).Side = StrikeOffEvents.ValidStrides(ptr).Side
            StrikeOffEvents.ValidStrikeOffEvents(cnt).Type = FootStrike
            StrikeOffEvents.ValidStrikeOffEvents(cnt).Time = StrikeOffEvents.ValidStrides(ptr).StartStrikeTime
            cnt = cnt + 1
            StrikeOffEvents.ValidStrikeOffEvents(cnt).Side = OtherSide(StrikeOffEvents.ValidStrides(ptr).Side)
            StrikeOffEvents.ValidStrikeOffEvents(cnt).Type = FootOff
            StrikeOffEvents.ValidStrikeOffEvents(cnt).Time = StrikeOffEvents.ValidStrides(ptr).OppositeFootOffTime
            cnt = cnt + 1
            StrikeOffEvents.ValidStrikeOffEvents(cnt).Side = OtherSide(StrikeOffEvents.ValidStrides(ptr).Side)
            StrikeOffEvents.ValidStrikeOffEvents(cnt).Type = FootStrike
            StrikeOffEvents.ValidStrikeOffEvents(cnt).Time = StrikeOffEvents.ValidStrides(ptr).OppositeFootStrikeTime
            cnt = cnt + 1
            StrikeOffEvents.ValidStrikeOffEvents(cnt).Side = StrikeOffEvents.ValidStrides(ptr).Side
            StrikeOffEvents.ValidStrikeOffEvents(cnt).Type = FootOff
            StrikeOffEvents.ValidStrikeOffEvents(cnt).Time = StrikeOffEvents.ValidStrides(ptr).SameFootOffTime
            cnt = cnt + 1
            StrikeOffEvents.ValidStrikeOffEvents(cnt).Side = StrikeOffEvents.ValidStrides(ptr).Side
            StrikeOffEvents.ValidStrikeOffEvents(cnt).Type = FootStrike
            StrikeOffEvents.ValidStrikeOffEvents(cnt).Time = StrikeOffEvents.ValidStrides(ptr).StopStrikeTime
        Next ptr
        ' Now we must eliminate duplicates from this ValidStrikeOffEvents array.
        cnt = 1
        Do
            If cnt >= StrikeOffEvents.ValidStrikeOffCount Then Exit Do
            DupeFound = False
            For ptr = cnt + 1 To StrikeOffEvents.ValidStrikeOffCount
                If StrikeOffEvents.ValidStrikeOffEvents(cnt).Side = StrikeOffEvents.ValidStrikeOffEvents(ptr).Side And _
                   StrikeOffEvents.ValidStrikeOffEvents(cnt).Type = StrikeOffEvents.ValidStrikeOffEvents(ptr).Type And _
                   StrikeOffEvents.ValidStrikeOffEvents(cnt).Time = StrikeOffEvents.ValidStrikeOffEvents(ptr).Time Then
                    DupeFound = True
                    Exit For
                End If
            Next ptr
            If DupeFound Then
                ' Duplicate found. Eliminate it.
                For ptr = cnt + 1 To StrikeOffEvents.ValidStrikeOffCount - 1
                    StrikeOffEvents.ValidStrikeOffEvents(ptr) = StrikeOffEvents.ValidStrikeOffEvents(ptr + 1)
                Next ptr
                StrikeOffEvents.ValidStrikeOffCount = StrikeOffEvents.ValidStrikeOffCount - 1
            Else
                cnt = cnt + 1
            End If
        Loop
        ' Redimension so that Ubound is accurate.
        ReDim Preserve StrikeOffEvents.ValidStrikeOffEvents(1 To StrikeOffEvents.ValidStrikeOffCount)
    Else ' Dim arrays to (0) if there is no data so that ubound can be used.
        ReDim StrikeOffEvents.ValidStrides(0)
        ReDim StrikeOffEvents.ValidStrikeOffEvents(0)
    End If
    '
    ' And we're done.
    Exit Sub
    '
ErrorInStrikeOffEvents:
    StrikeOffEvents = BlankStrikeOffEvents
    Exit Sub
End Sub

Private Sub PrepareStrikeOffEventFramesArray()
    ' Even though this is public, it should be called only from the OpenC3dFile procedure.
    Dim FramePtr As Long
    Dim EventPtr As Long
    Dim TempTime As Single
    '
    ' Get strike/off events.
    StrikeOffEventFramesCount = StrikeOffEvents.ValidStrikeOffCount
    '
    ' If there aren't any valid strike/off events, then we're done.
    If StrikeOffEventFramesCount = 0 Then
        ReDim StrikeOffEventFrames(0)
        Exit Sub
    End If
    '
    ' Found strike/off events, so we've got some work to do.
    ReDim StrikeOffEventFrames(1 To StrikeOffEventFramesCount)
    For EventPtr = 1 To StrikeOffEventFramesCount
        TempTime = StrikeOffEvents.ValidStrikeOffEvents(EventPtr).Time
        FramePtr = Round(TempTime / C3dHeader.TimeBetweenFrames) + 1
        StrikeOffEventFrames(EventPtr).FrameNumber = FramePtr - C3dHeader.FirstPointFrameNumber + 1
        StrikeOffEventFrames(EventPtr).StrideEventSide = StrikeOffEvents.ValidStrikeOffEvents(EventPtr).Side
        StrikeOffEventFrames(EventPtr).StrideEventType = StrikeOffEvents.ValidStrikeOffEvents(EventPtr).Type
    Next EventPtr
End Sub

Private Sub SortStrikeOffEventGroup(Count As Long, Events() As StrikeOffSingleEventType)
    ' This is used only in the SetStrikeOffEvents procedure.
    Dim Sort1 As Long
    Dim Sort2 As Long
    Dim TempEvent As StrikeOffSingleEventType
    '
    For Sort1 = 1 To Count - 1
        For Sort2 = Sort1 + 1 To Count
            If Events(Sort1).Time > Events(Sort2).Time Then
                TempEvent = Events(Sort1)
                Events(Sort1) = Events(Sort2)
                Events(Sort2) = TempEvent
            End If
        Next Sort2
    Next Sort1
End Sub

Private Function OtherSide(TheSide As FootSideEnum) As FootSideEnum
    Select Case TheSide
    Case LeftSide
        OtherSide = RightSide
    Case RightSide
        OtherSide = LeftSide
    Case Else
        OtherSide = NoSide
    End Select
End Function



